"""Jedi types mapped to LSP types"""

from typing import Optional
from pygls.types import CompletionItemKind

# Taken from: https://github.com/palantir/python-language-server
_JEDI_LSP_TYPE_MAP = {
    "buffer": CompletionItemKind.Class,
    "builtin": CompletionItemKind.Class,
    "builtinfunction": CompletionItemKind.Function,
    "class": CompletionItemKind.Class,
    "constant": CompletionItemKind.Variable,
    "dict": CompletionItemKind.Class,
    "dictionary": CompletionItemKind.Class,
    "dictproxy": CompletionItemKind.Class,
    "file": CompletionItemKind.File,
    "frame": CompletionItemKind.Class,
    "funcdef": CompletionItemKind.Function,
    "function": CompletionItemKind.Function,
    "generator": CompletionItemKind.Function,
    "import": CompletionItemKind.Module,
    "instance": CompletionItemKind.Reference,
    "keyword": CompletionItemKind.Keyword,
    "lambda": CompletionItemKind.Function,
    "method": CompletionItemKind.Method,
    "module": CompletionItemKind.Module,
    "none": CompletionItemKind.Value,
    "param": CompletionItemKind.Variable,
    "property": CompletionItemKind.Property,
    "slice": CompletionItemKind.Class,
    "statement": CompletionItemKind.Keyword,
    "traceback": CompletionItemKind.Class,
    "tuple": CompletionItemKind.Class,
    "type": CompletionItemKind.Class,
    "value": CompletionItemKind.Value,
    "variable": CompletionItemKind.Variable,
    "xrange": CompletionItemKind.Class,
}


def get_lsp_type(jedi_type: str) -> Optional[int]:
    """Get type map"""
    return _JEDI_LSP_TYPE_MAP.get(jedi_type)



"""Jedi Language Server

Creates the language server constant and wraps "features" with it.

Official language server spec:
    https://microsoft.github.io/language-server-protocol/specification
"""

from typing import List, Optional

from pygls.server import LanguageServer
from pygls.features import (
    COMPLETION,
    DEFINITION,
    HOVER,
    REFERENCES,
    RENAME,
    TEXT_DOCUMENT_DID_CHANGE,
    TEXT_DOCUMENT_DID_OPEN,
)
from pygls.types import (
    CompletionItem,
    CompletionList,
    CompletionParams,
    DidChangeTextDocumentParams,
    DidOpenTextDocumentParams,
    Hover,
    Location,
    RenameParams,
    TextDocumentPositionParams,
    TextDocumentSyncKind,
    TextEdit,
    WorkspaceEdit,
)

from .type_map import get_lsp_type

from .server_utils import get_jedi_script, locations_from_definitions


SERVER = LanguageServer()

# NOTE: neovim LanguageClient doesn't support incremental completions. When it
# does, I may consider relaxing this constraint
SERVER.sync_kind = TextDocumentSyncKind.FULL


@SERVER.feature(COMPLETION, triggerCharacters=["."])
def lsp_completion(server: LanguageServer, params: CompletionParams = None):
    """Returns completion items."""
    script = get_jedi_script(server, params)
    jedi_completions = script.completions()
    return CompletionList(
        is_incomplete=False,
        items=[
            CompletionItem(
                label=completion.name,
                kind=get_lsp_type(completion.type),
                detail=completion.description,
                documentation=completion.docstring(),
                insert_text=completion.name,
            )
            for completion in jedi_completions
        ],
    )


@SERVER.feature(DEFINITION)
def lsp_definition(
    server: LanguageServer, params: TextDocumentPositionParams
) -> List[Location]:
    """Support Goto Definition"""
    script = get_jedi_script(server, params)
    definitions = script.goto_assignments(
        follow_imports=True, follow_builtin_imports=True
    )
    return locations_from_definitions(definitions)


@SERVER.feature(HOVER)
def lsp_hover(
    server: LanguageServer, params: TextDocumentPositionParams
) -> Hover:
    """Support the hover feature"""
    script = get_jedi_script(server, params)
    definitions = script.goto_definitions()
    return Hover(
        contents=(
            definitions[0].docstring()
            if definitions
            else "No docstring definition found."
        )
    )


@SERVER.feature(REFERENCES)
def lsp_references(
    server: LanguageServer, params: TextDocumentPositionParams
) -> List[Location]:
    """Obtain all references to document"""
    script = get_jedi_script(server, params)
    definitions = script.usages()
    return locations_from_definitions(definitions)


@SERVER.feature(RENAME)
def lsp_rename(
    server: LanguageServer, params: RenameParams
) -> Optional[WorkspaceEdit]:
    """Optional workspace edit"""
    script = get_jedi_script(server, params)
    definitions = script.usages()
    locations = locations_from_definitions(definitions)
    if not locations:
        return None
    changes = {}
    for location in locations:
        text_edit = TextEdit(location.range, new_text=params.newName)
        if location.uri not in changes:
            changes[location.uri] = [text_edit]
        else:
            changes[location.uri].append(text_edit)
    return WorkspaceEdit(changes=changes)


@SERVER.feature(TEXT_DOCUMENT_DID_CHANGE)
def lsp_did_change(
    server: LanguageServer, params: DidChangeTextDocumentParams
) -> None:
    """Updates file when the file changes on server"""
    workspace = server.workspace
    if params.contentChanges:
        for change in params.contentChanges:
            workspace.update_document(
                text_doc=params.textDocument, change=change
            )


@SERVER.feature(TEXT_DOCUMENT_DID_OPEN)
def lsp_did_open(
    server: LanguageServer, params: DidOpenTextDocumentParams
) -> None:
    """Text document did, in fact, open"""
    workspace = server.workspace
    workspace.put_document(params.textDocument)



"""Utility functions used by the language server"""

from typing import Union, List

from jedi import Script
from jedi.api.classes import Definition

from pygls.server import LanguageServer
from pygls.uris import from_fs_path
from pygls.types import (
    Location,
    Range,
    Position,
    RenameParams,
    TextDocumentPositionParams,
)


def get_jedi_script(
    server: LanguageServer,
    params: Union[TextDocumentPositionParams, RenameParams],
) -> Script:
    """Simplifies getting jedi Script

    NOTE:
        * jedi is 1-indexed for lines and 0-indexed for columns
        * LSP is 0-indexed for lines and 0-indexed for columns
        * Therefore, add 1 to LSP's request for the line
    """
    workspace = server.workspace
    text_doc = workspace.get_document(params.textDocument.uri)
    return Script(
        source=text_doc.source,
        path=text_doc.path,
        line=params.position.line + 1,
        column=params.position.character,
    )


def locations_from_definitions(
    definitions: List[Definition]
) -> List[Location]:
    """Get LSP locations from Jedi definitions

    NOTE:
        * jedi is 1-indexed for lines and 0-indexed for columns
        * LSP is 0-indexed for lines and 0-indexed for columns
        * Therefore, subtract 1 from Jedi's definition line
    """
    return [
        Location(
            uri=from_fs_path(definition.module_path),
            range=Range(
                start=Position(
                    line=definition.line - 1, character=definition.column
                ),
                end=Position(
                    line=definition.line - 1,
                    character=definition.column + len(definition.name),
                ),
            ),
        )
        for definition in definitions
    ]



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['jedi_language_server']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.0,<8.0', 'jedi>=0.15.1,<0.16.0', 'pygls>=0.8.0,<0.9.0']

entry_points = \
{'console_scripts': ['jedi-language-server = jedi_language_server.cli:cli']}

setup_kwargs = {
    'name': 'jedi-language-server',
    'version': '0.1.0',
    'description': 'Toml sorting library',
    'long_description': '# jedi-language-server\n\n[![image-version](https://img.shields.io/pypi/v/jedi-language-server.svg)](https://python.org/pypi/jedi-language-server)\n[![image-license](https://img.shields.io/pypi/l/jedi-language-server.svg)](https://python.org/pypi/jedi-language-server)\n[![image-python-versions](https://img.shields.io/pypi/pyversions/jedi-language-server.svg)](https://python.org/pypi/jedi-language-server)\n\nA [Language Server](https://microsoft.github.io/language-server-protocol/) for [Jedi](https://jedi.readthedocs.io/en/latest/).\n\n## Installation\n\n```bash\n# With pip\npip install jedi-language-server\n\n# With poetry\npoetry add jedi-language-server\n```\n\n## Motivation\n\nMy motivation goes here...\n\n## Usage\n\nIt currently works only over IO.\n\n```bash\njedi-language-server\n```\n\n## Local Development\n\nLocal development for this project is quite simple.\n\n**Dependencies**\n\nInstall the following tools manually.\n\n* [Poetry](https://github.com/sdispater/poetry#installation)\n* [GNU Make](https://www.gnu.org/software/make/)\n\n*Recommended*\n\n* [pyenv](https://github.com/pyenv/pyenv)\n\n**Set up development environment**\n\n```bash\nmake setup\n```\n\n**Run Tests**\n\n```bash\nmake test\n```\n\n## Written by\n\nSamuel Roeca *samuel.roeca@gmail.com*\n',
    'author': 'Sam Roeca',
    'author_email': 'samuel.roeca@gmail.com',
    'url': 'https://github.com/pappasam/jedi-language-server',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
